//+------------------------------------------------------------------+
//|                                                       OpenCL.mqh |
//|                        Copyright 2012, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Simple                                                           |
//+------------------------------------------------------------------+
class COpenCL
  {
private:
   int               m_context;             // handle of OpenCL object
   int               m_program;             // handle of OpenCL program
   int               m_kernel;              // handle of OpenCL kernel
   int               m_buffer;              // handle of OpenCL buffer

public:

                     COpenCL(void):m_context(INVALID_HANDLE),m_program(INVALID_HANDLE),m_kernel(INVALID_HANDLE)
     {
     }

                    ~COpenCL(void)
     {
      Free();
     }

   //--- methods for working with OpenCL object
   bool              Create(const string source,const string kernel,const bool use_cpu=false)
     {
      //--- create a context
      if(m_context==INVALID_HANDLE)
        {
         m_context=CLContextCreate(use_cpu);
         if(m_context==INVALID_HANDLE)
           {
            Print("COpenCL: create context failed");
            return(false);
           }
        }
      //--- delete existing program
      if(m_buffer!=INVALID_HANDLE)
        {
         CLBufferFree(m_buffer);
         m_buffer=INVALID_HANDLE;
        }
      if(m_kernel!=INVALID_HANDLE)
        {
         CLKernelFree(m_kernel);
         m_kernel=INVALID_HANDLE;
        }
      if(m_program!=INVALID_HANDLE)
        {
         CLProgramFree(m_program);
         m_program=INVALID_HANDLE;
        }
      //--- create a program
      m_program=CLProgramCreate(m_context,source);
      if(m_program==INVALID_HANDLE)
        {
         Free();
         Print("COpenCL: create program failed, check source code");
         return(false);
        }
      //--- create a kernel
      m_kernel=CLKernelCreate(m_program,kernel);
      if(m_kernel==INVALID_HANDLE)
        {
         Free();
         Print("COpenCL: create kernel failed");
         return(false);
        }
      //---
      return(true);
     }

   //---
   bool CreateBuffer(const uint size,const uint flags)
     {
      //---
      if(m_kernel!=INVALID_HANDLE)
        {
         //--- delete
         if(m_buffer!=INVALID_HANDLE)
            CLBufferFree(m_buffer);
         //--- create new
         m_buffer=CLBufferCreate(m_context,size,flags);
         if(m_buffer!=INVALID_HANDLE)
            return(true);
        }
      //---
      Print("COpenCL: create buffer failed, ",size);
      return(false);
     }

   //---
   void              Free(void)
     {
      //---
      if(m_buffer!=INVALID_HANDLE)
        {
         CLBufferFree(m_buffer);
         m_buffer=INVALID_HANDLE;
        }
      //---
      if(m_kernel!=INVALID_HANDLE)
        {
         CLKernelFree(m_kernel);
         m_kernel=INVALID_HANDLE;
        }
      //---
      if(m_program!=INVALID_HANDLE)
        {
         CLProgramFree(m_program);
         m_program=INVALID_HANDLE;
        }
      //---
      if(m_context!=INVALID_HANDLE)
        {
         CLContextFree(m_context);
         m_context=INVALID_HANDLE;
        }
     }

   //---
   bool SetArg(const uint arg_index,const MqlParam &param)
     {
      if(m_kernel!=INVALID_HANDLE)
         switch(param.type)
           {
            case TYPE_BOOL:
               Print("COpenCL: unsupported TYPE_BOOL type in SetArg");
               break;

            case TYPE_CHAR:
               return CLSetKernelArg(m_kernel,arg_index,(char)param.integer_value);

            case TYPE_UCHAR:
               return CLSetKernelArg(m_kernel,arg_index,(uchar)param.integer_value);

            case TYPE_SHORT:
               return CLSetKernelArg(m_kernel,arg_index,(short)param.integer_value);

            case TYPE_USHORT:
               return CLSetKernelArg(m_kernel,arg_index,(ushort)param.integer_value);

            case TYPE_COLOR:
               return CLSetKernelArg(m_kernel,arg_index,(color)param.integer_value);

            case TYPE_INT:
               return CLSetKernelArg(m_kernel,arg_index,(int)param.integer_value);

            case TYPE_UINT:
               return CLSetKernelArg(m_kernel,arg_index,(uint)param.integer_value);

            case TYPE_DATETIME:
               return CLSetKernelArg(m_kernel,arg_index,(datetime)param.integer_value);

            case TYPE_LONG:
               return CLSetKernelArg(m_kernel,arg_index,param.integer_value);

            case TYPE_ULONG:
               return CLSetKernelArg(m_kernel,arg_index,(ulong)param.integer_value);

            case TYPE_FLOAT:
               return CLSetKernelArg(m_kernel,arg_index,(float)param.double_value);

            case TYPE_DOUBLE:
               return CLSetKernelArg(m_kernel,arg_index,param.double_value);

            case TYPE_STRING:
               Print("COpenCL: unsupported TYPE_STRING type in SetArg");
               break;

            default:
               Print("COpenCL: unknown type in SetArg");
           }
      //---
      return(false);
     }

   //---
   bool SetArgMemBuffer(const uint arg_index)
     {
      //---
      if(m_kernel!=INVALID_HANDLE)
         if(m_buffer!=INVALID_HANDLE)
            return CLSetKernelArgMem(m_kernel,arg_index,m_buffer);
      //---
      return(false);
     }

   //---
   bool SetArgList(const uint parameters_cnt,const MqlParam &parameters_array[])
     {
      //---
      for(uint i=0;i<parameters_cnt;i++)
         if(!SetArg(i,parameters_array[i]))
            return(false);
      //---
      return(true);
     }

   //---
   bool WriteBool(const bool &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadBool(bool &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteChar(const char &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadChar(char &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteUchar(const uchar &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadUchar(uchar &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteShort(const short &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadShort(short &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteUshort(const ushort &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadUshort(ushort &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteColor(const color &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadColor(color &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteInt(const int &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadInt(int &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteUint(const uint &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadUint(uint &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteDatetime(const datetime &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadDatetime(datetime &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteLong(const long &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadLong(long &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteUlong(const ulong &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadUlong(ulong &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteFloat(const float &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadFloat(float &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteDouble(const double &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadDouble(double &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   bool WriteString(const string &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferWrite(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }
   bool ReadString(string &data[],uint buffer_offset=0,uint array_offset=0,uint array_count=WHOLE_ARRAY)
     {
      return(CLBufferRead(m_buffer,data,buffer_offset,array_offset,array_count)!=0);
     }

   //---
   bool Execute(const uint work_dim,const uint &global_work_offset[],const uint &global_work_size[])
     {
      //---
      if(m_context!=INVALID_HANDLE)
         if(m_kernel!=INVALID_HANDLE)
            return CLExecute(m_kernel,work_dim,global_work_offset,global_work_size);
      //---
      return(false);
     }
  };
//+------------------------------------------------------------------+
